# -*- coding: utf-8 -*-

"""Parse GPT-3 output to produce Dict[Formula, Float] result."""
import re
from dataclasses import dataclass
from typing import Dict, Match, Set, Tuple, cast

from pylogics.syntax.base import Formula

from nl2ltl.engines.utils import _get_formulas
from nl2ltl.filters.base import Filter


@dataclass
class GPT3Output:
    """Dataclass to represent the GPT-3 output."""

    pattern: str
    entities: Tuple[str]

    def __post_init__(self):
        """Do consistency checks after initialization."""
        assert self.pattern is not None
        assert self.entities is not None


@dataclass
class _GPT3OutputWrapper:
    """A wrapper to the textual output of GPT-3."""

    output: dict

    @property
    def pattern(self) -> str:
        """Get the predicted pattern."""
        return str(
            cast(
                Match, re.search("PATTERN: (.*)\n", self.output["choices"][0]["text"])
            ).group(1)
        )

    @property
    def entities(self) -> Tuple[str]:
        """Get the predicted entities."""
        return tuple(
            cast(
                Match,
                re.search("SYMBOLS: (.*)", self.output["choices"][0]["text"]),
            )
            .group(1)
            .split(", ")
        )


def parse_gpt3_output(gpt3_output: dict) -> GPT3Output:
    """
    Parse the GPT-3 output.

    :param gpt3_output: the json description of the GPT-3 prediction.
    :return: a GPT3Output instance.
    """
    wrapper = _GPT3OutputWrapper(gpt3_output)
    pattern: str = wrapper.pattern
    entities: Tuple[str] = wrapper.entities
    gpt3_result = GPT3Output(pattern, entities)
    return gpt3_result


def parse_gpt3_result(
    output: GPT3Output, filtering: Filter = None
) -> Dict[Formula, float]:
    """
    Build a dict of formulas, given the GPT3Output object.

    :param output: a GPT3Output instance.
    :param filtering: a custom filtering function
    :return: the dictionary of formulas.
    """
    result: Dict[Formula, float] = {}
    symbols: Dict[str, float] = {e: 1 for e in output.entities}
    formulas: Set[Formula] = _get_formulas(output.pattern, symbols)
    if all(isinstance(f, Formula) for f in formulas):
        for f in formulas:
            result[f] = 1
    else:
        raise Exception("The output is not a valid formula.")

    if filtering:
        return filtering.enforce(result, symbols)
    else:
        return result



# -*- coding: utf-8 -*-

"""
Implementation of the RASA engine.

Repository:

    https://github.com/RasaHQ/rasa

"""
import asyncio
import shutil
from pathlib import Path
from typing import Dict

import rasa
from pylogics.syntax.base import Formula
from rasa.core.agent import Agent

from nl2ltl.engines.base import Engine
from nl2ltl.engines.rasa import ENGINE_ROOT
from nl2ltl.engines.rasa.helpers import _get_latest_model
from nl2ltl.engines.rasa.output import RasaOutput, parse_rasa_output, parse_rasa_result
from nl2ltl.filters.base import Filter

engine_root = ENGINE_ROOT
DATA_DIR = engine_root / "data"
DOMAIN_PATH = engine_root / DATA_DIR / "domain.yml"
CONFIG_PATH = engine_root / DATA_DIR / "config.yml"
TRAINING_PATH = engine_root / DATA_DIR / "nlu_training.yml"
MODEL_OUTPUT_PATH = engine_root / "models"


class RasaEngine(Engine):
    """The RASA engine."""

    def __init__(
        self,
        model: Path = None,
    ):
        """Rasa NLU Engine initialization."""
        self._load_model(model)

    def _load_model(self, model: Path = None):
        if model:
            self.agent = Agent.load(model)
        else:
            self.agent = Agent.load(_get_latest_model(path=MODEL_OUTPUT_PATH))

    @classmethod
    def __check_rasa_available(cls):
        """Check that the Rasa tool is available."""
        is_rasa_present = shutil.which("rasa") is not None
        if is_rasa_present is None:
            raise Exception(
                "Rasa is not installed. Please follow"
                "the installation instructions at https://rasa.com/docs/rasa/installation.\n"
            )

    @classmethod
    def __check_rasa_version(cls):
        """Check that the Rasa tool is at the right version."""
        is_right_version = rasa.__version__ == "3.2.2"
        if not is_right_version:
            raise Exception(
                "Rasa needs to be at version 3.2.2. "
                "Please install it manually using:"
                "\n"
                "pip install rasa==3.2.2"
            )

    def __post_init__(self):
        """Do post-initialization checks."""
        self.__check_rasa_available()
        self.__check_rasa_version()

    @staticmethod
    def train(
        domain: Path = DOMAIN_PATH,
        config: Path = CONFIG_PATH,
        training: Path = TRAINING_PATH,
    ) -> Path:
        """Train a Rasa model."""
        result = rasa.train(
            domain=str(domain),
            config=str(config),
            training_files=[str(training)],
            output=str(MODEL_OUTPUT_PATH),
            force_training=True,
        )
        return Path(result.model)

    def translate(
        self, utterance: str, filtering: Filter = None
    ) -> Dict[Formula, float]:
        """From NL to best matching LTL formulas with confidence."""
        return _process_utterance(utterance, self.agent, filtering)


def _process_utterance(
    utterance: str, rasa_agent: Agent, filtering: Filter
) -> Dict[Formula, float]:
    """
    Process NL utterance.

    :param utterance: the natural language utterance
    :return: a dict with matching formulas and confidence
    """
    prediction = asyncio.run(rasa_agent.parse_message(utterance.strip()))
    rasa_result: RasaOutput = parse_rasa_output(prediction)
    matching_formulas: Dict[Formula, float] = parse_rasa_result(rasa_result, filtering)
    return matching_formulas



# -*- coding: utf-8 -*-

"""
Implementation of the GPT-3 engine.

Website:

    https://openai.com

"""
import json
import os
from pathlib import Path
from typing import Dict

import openai
from pylogics.syntax.base import Formula

from nl2ltl.engines.base import Engine
from nl2ltl.engines.gpt3 import ENGINE_ROOT
from nl2ltl.engines.gpt3.output import GPT3Output, parse_gpt3_output, parse_gpt3_result
from nl2ltl.filters.base import Filter

openai.api_key = os.getenv("OPENAI_API_KEY")
engine_root = ENGINE_ROOT
DATA_DIR = engine_root / "data"
PROMPT_PATH = engine_root / DATA_DIR / "prompt.json"


class GPT3Engine(Engine):
    """The GPT-3 engine."""

    def __init__(
        self,
        prompt: Path = PROMPT_PATH,
        temperature: float = 0.5,
    ):
        """GPT-3 LLM Engine initialization."""
        self.prompt = self._load_prompt(prompt)
        self.temperature = temperature

    def _load_prompt(self, prompt):
        return json.load(open(prompt, "r"))["prompt"]

    @classmethod
    def __check_openai_version(cls):
        """Check that the GPT-3 tool is at the right version."""
        is_right_version = openai.__version__ == "0.23.0"
        if not is_right_version:
            raise Exception(
                "OpenAI needs to be at version 0.23.0. "
                "Please install it manually using:"
                "\n"
                "pip install openai==0.23.0"
            )

    def __post_init__(self):
        """Do post-initialization checks."""
        self.__check_openai_version()

    def translate(
        self, utterance: str, filtering: Filter = None
    ) -> Dict[Formula, float]:
        """From NL to best matching LTL formulas with confidence."""
        return _process_utterance(utterance, self.prompt, self.temperature, filtering)


def _process_utterance(
    utterance: str, prompt: str, temperature: float, filtering: Filter
) -> Dict[Formula, float]:
    """
    Process NL utterance.

    :param utterance: the natural language utterance
    :return: a dict with matching formulas and confidence
    """
    query = "NL: " + utterance + "\n"
    prediction = openai.Completion.create(
        model="text-davinci-002",
        prompt=prompt + query,
        temperature=temperature,
        max_tokens=200,
        top_p=1.0,
        frequency_penalty=0.0,
        presence_penalty=0.0,
        stop=["\n\n"],
    )
    gpt3_result: GPT3Output = parse_gpt3_output(prediction)
    matching_formulas: Dict[Formula, float] = parse_gpt3_result(gpt3_result, filtering)
    return matching_formulas



# -*- coding: utf-8 -*-

"""Implementation of the GPT-3 engine."""

import inspect
from pathlib import Path

ENGINE_ROOT = Path(inspect.getframeinfo(inspect.currentframe()).filename).parent  # type: ignore



# -*- coding: utf-8 -*-

"""Engines utils."""
import difflib
from typing import Callable, Dict, Set

from pylogics.syntax.base import Formula

from nl2ltl.declare.base import TemplateEnum


def _get_formulas(name: str, args: Dict[str, float]) -> Set[Formula]:
    """Instantiate matching formulas based on intent name and entities."""
    import nl2ltl.declare.declare
    import nl2ltl.engines.grounding

    class_name_match = difflib.get_close_matches(
        name, [x.value for x in TemplateEnum], n=1
    )
    grounding_map: Dict[str, Callable] = {}
    for c_name in TemplateEnum:
        grounding_map[c_name.value] = getattr(
            nl2ltl.engines.grounding, f"ground_{c_name.value.lower()}"
        )

    grounding_func: Callable = grounding_map[str(class_name_match[0])]
    grounded_formulas: Set[Formula] = grounding_func(args)
    return grounded_formulas


def pretty(result: Dict[Formula, float]):
    """Pretty print Rasa output."""
    print("=" * 150)
    for k, v in result.items():
        print(f"Declare Template: {str(k)}", end="\n")
        print(f"English meaning:  {k.to_english()}", end="\n")
        print(f"Confidence:       {str(v)}", end="\n\n")
